'use strict';

var assert = require('assert');
var Devebot = require('devebot');
var Promise = Devebot.require('bluebird');
var lodash = Devebot.require('lodash');
var pinbug = Devebot.require('pinbug');
var debugx = pinbug('devebot:co:rabbitmq:handler');
var debug0 = pinbug('trace:devebot:co:rabbitmq:handler');
var debug2 = pinbug('devebot:co:rabbitmq:handler:extract');
var amqp = require('amqplib/callback_api');
var locks = require('locks');
var util = require('util');
var Readable = require('stream').Readable;
var Writable = require('stream').Writable;
var zapper = require('./zapper');

var Handler = function Handler(params) {
  debugx.enabled && debugx(' + constructor begin ...');

  var self = this;
  var LX = this.logger || zapper.getLogger();
  var LT = this.tracer || zapper.getTracer();

  var config = {};
  config.uri = params.uri || params.host;
  config.exchangeName = params.exchangeName || params.exchange;
  config.exchangeType = params.exchangeType || 'fanout';
  if (zapper.isNumber(params.exchangeQuota)) {
    config.exchangeQuota = params.exchangeQuota;
  }
  if (zapper.isBoolean(params.exchangeMutex)) {
    config.exchangeMutex = params.exchangeMutex;
  }
  config.durable = zapper.isBoolean(params.durable) ? params.durable : true;
  config.autoDelete = zapper.isBoolean(params.autoDelete) ? params.autoDelete : false;
  if (zapper.isString(params.alternateExchange)) {
    config.alternateExchange = params.alternateExchange;
  }
  config.routingKey = zapper.isString(params.routingKey) ? params.routingKey : '';

  var originalInbox = params.inbox || params.consumer;
  config.inbox = { uri: config.uri };
  if (lodash.isObject(originalInbox)) {
    lodash.merge(config.inbox, originalInbox);
  }

  if (!zapper.isString(config.inbox.queueName)) {
    config.inbox.queueName = params.queueName || params.queue || '';
  }
  if (!zapper.isBoolean(config.inbox.durable)) {
    config.inbox.durable = zapper.isBoolean(params.durable) ? params.durable : true;
  }
  if (!zapper.isBoolean(config.inbox.exclusive)) {
    config.inbox.exclusive = zapper.isBoolean(params.exclusive) ? params.exclusive : false;
  }
  if (!zapper.isBoolean(config.inbox.noAck)) {
    config.inbox.noAck = zapper.isBoolean(params.noAck) ? params.noAck : false;
  }
  if (!zapper.isNumber(config.inbox.prefetch)) {
    config.inbox.prefetch = zapper.isNumber(params.prefetch) ? params.prefetch : undefined;
  }
  if (!zapper.isBoolean(config.inbox.connectIsCached)) {
    config.inbox.connectIsCached = zapper.isBoolean(params.connectIsCached) ? params.connectIsCached : true;
  }
  if (!zapper.isBoolean(config.inbox.channelIsCached)) {
    config.inbox.channelIsCached = zapper.isBoolean(params.channelIsCached) ? params.channelIsCached : true;
  }
  if (!zapper.isBoolean(config.inbox.noBinding)) {
    config.inbox.noBinding = false;
  }
  if (!zapper.isNumber(config.inbox.maxSubscribers)) {
    config.inbox.maxSubscribers = params.maxSubscribers;
  }

  var originalTrash = params.trash || params.recycler;
  if (lodash.isObject(originalTrash)) {
    config.trash = {
      uri: config.uri,
      redeliveredCountName: 'x-redelivered-count',
      redeliveredLimit: 5
    };
    lodash.merge(config.trash, originalTrash);
    if (lodash.isEmpty(config.trash.queueName)) {
      config.trash.queueName = config.inbox.queueName + '-trashed';
    }
  }

  config.maxListeners = 20;
  if (zapper.isPositiveInteger(params.maxListeners)) {
    config.maxListeners = params.maxListeners;
  }
  if (config.inbox) {
    config.inbox.maxListeners = config.inbox.maxListeners || config.maxListeners;
  }
  if (config.chain) {
    config.chain.maxListeners = config.chain.maxListeners || config.maxListeners;
  }
  if (config.trash) {
    config.trash.maxListeners = config.trash.maxListeners || config.maxListeners;
  }

  debugx.enabled && debugx(' - configuration object: %s', JSON.stringify(config));

  Object.defineProperty(this, 'config', {
    get: function get() {
      return config;
    },
    set: function set(value) {}
  });

  debugx.enabled && debugx(' - constructor end!');
};

var getTicket = function getTicket(kwargs) {
  var state = kwargs.state;

  var ticket = void 0;
  if (state.guard) {
    ticket = new Promise(function (onResolved, onRejected) {
      state.guard.lock(function () {
        onResolved();
      });
    });
    ticket.finally(function () {
      state.guard.unlock();
    });
  } else {
    ticket = Promise.resolve();
  }
  return ticket;
};

var getConnection = function getConnection(kwargs) {
  var state = kwargs.state,
      config = kwargs.config;


  state.connectionCount = state.connectionCount || 0;
  debugx.enabled && debugx('getConnection() - connection amount: %s', state.connectionCount);

  if (config.connectIsCached !== false && state.connection) {
    debugx.enabled && debugx('getConnection() - connection has been available');
    return Promise.resolve(state.connection);
  } else {
    debugx.enabled && debugx('getConnection() - make a new connection');
    var amqp_connect = Promise.promisify(amqp.connect, { context: amqp });
    return amqp_connect(config.uri, {}).then(function (conn) {
      state.connectionCount += 1;
      if (config.maxListeners > 0) {
        conn.setMaxListeners(config.maxListeners);
      }
      conn.on('close', function () {
        debugx.enabled && debugx('getConnection() - connection is closed');
        delete state.connection;
        state.connectionCount--;
      });
      conn.on('error', function (err) {
        debugx.enabled && debugx('getConnection() - connection has error');
        delete state.connection;
      });
      debugx.enabled && debugx('getConnection() - connection is created successfully');
      return state.connection = conn;
    }).catch(function (err) {
      return Promise.resolve(kwargs).delay(1100).then(getConnection);
    });
  }
};

var getChannel = function getChannel(kwargs) {
  var state = kwargs.state,
      config = kwargs.config;

  if (config.channelIsCached !== false && state.channel) {
    debugx.enabled && debugx('getChannel() - channel has been available');
    return Promise.resolve(state.channel);
  }
  return getTicket(kwargs).then(function () {
    if (config.channelIsCached !== false && state.channel) {
      debugx.enabled && debugx('getChannel() - channel has been available');
      return Promise.resolve(state.channel);
    } else {
      debugx.enabled && debugx('getChannel() - make a new channel');
      return getConnection(kwargs).then(function (conn) {
        debugx.enabled && debugx('getChannel() - connection has already');
        var createChannel = Promise.promisify(conn.createConfirmChannel, { context: conn });
        return createChannel().then(function (ch) {
          ch.on('close', function () {
            debugx.enabled && debugx('getChannel() - channel is closed');
            delete state.channel;
          });
          ch.on('error', function (err) {
            debugx.enabled && debugx('getChannel() - channel has error');
            delete state.channel;
          });
          ch.on('drain', function () {
            debugx.enabled && debugx('getChannel() - channel has drained, open the valve');
            state.valve && state.valve.set(true);
          });
          debugx.enabled && debugx('getChannel() - channel is created successfully');
          return state.channel = ch;
        }).catch(function (err) {
          return Promise.resolve(kwargs).delay(1700).then(getChannel);
        });
      });
    }
  });
};

var retrieveExchange = function retrieveExchange(kwargs, override) {
  var config = kwargs.config;

  override = override || {};
  return getChannel(kwargs).then(function (ch) {
    var ok = assertExchange(kwargs, ch);
    return ok.then(function (eok) {
      return {
        exchangeName: eok.exchange || config.exchangeName,
        routingKey: override.routingKey || config.routingKey
      };
    });
  });
};

var assertExchange = function assertExchange(kwargs, ch) {
  var state = kwargs.state,
      config = kwargs.config;

  if (state.exchangeAsserted) return Promise.resolve(state.exchangeAsserted);
  var ch_assertExchange = Promise.promisify(ch.assertExchange, { context: ch });
  return ch_assertExchange(config.exchangeName, config.exchangeType, {
    durable: config.durable,
    autoDelete: config.autoDelete
  }).then(function (eok) {
    return state.exchangeAsserted = eok;
  });
};

var assertQueue = function assertQueue(kwargs, ch, forced) {
  var state = kwargs.state,
      config = kwargs.config;

  if (!forced && state.queueAsserted) return Promise.resolve(state.queueAsserted);
  return Promise.promisify(ch.assertQueue, { context: ch })(config.queueName, {
    durable: config.durable,
    exclusive: config.exclusive
  }).then(function (qok) {
    return state.queueAsserted = qok;
  });
};

var checkQueue = function checkQueue(kwargs, ch) {
  return Promise.promisify(ch.checkQueue, { context: ch })(kwargs.config.queueName);
};

var purgeQueue = function purgeQueue(kwargs, ch) {
  return Promise.promisify(ch.purgeQueue, { context: ch })(kwargs.config.queueName);
};

var sendToQueue = function sendToQueue(kwargs, data, opts) {
  var state = kwargs.state,
      config = kwargs.config;

  opts = opts || {};
  debugx.enabled && debugx('%s() an object to rabbitmq queue', kwargs.name);
  return getChannel(kwargs).then(function (ch) {
    return assertQueue(kwargs, ch).then(function (qok) {
      var connection = state.connection;
      return new Promise(function (onResolved, onRejected) {
        var unexpectedClosing = function unexpectedClosing() {
          onRejected({ msg: 'Timeout exception' });
        };
        connection.on('error', unexpectedClosing);
        try {
          checkValve(kwargs, function () {
            var drained = ch.sendToQueue(config.queueName, zapper.bufferify(data), opts, function (err, ok) {
              connection.removeListener('error', unexpectedClosing);
              if (err) {
                onRejected(err);
              } else {
                ok = ok || {};
                ok.drained = true;
                onResolved(ok);
              }
            });
            if (!drained) {
              debugx.enabled && debugx('channel is overflowed, close the valve');
              state.valve && state.valve.set(false);
            }
          });
        } catch (exception) {
          connection.removeListener('error', unexpectedClosing);
          onRejected(exception);
        }
      });
    });
  }).catch(function (err) {
    debugx.enabled && debugx('sendToQueue() has failed, recall ...');
    state.connection = null;
    state.channel = null;
    return sendToQueue(kwargs, data, opts);
  });
};

var checkValve = function checkValve(ctx, whenConditionMet) {
  var state = ctx.state;

  if (state.valve) {
    state.valve.wait(function conditionTest(value) {
      return value === true;
    }, whenConditionMet);
  } else {
    whenConditionMet();
  }
};

var sendToExchange = function sendToExchange(ctx, data, opts, target) {
  return getChannel(ctx).then(function (channel) {
    return new Promise(function (onResolved, onRejected) {
      var unexpectedClosing = function unexpectedClosing() {
        onRejected({ msg: 'Timeout exception' });
      };
      var connection = ctx.state.connection;
      connection.on('error', unexpectedClosing);
      try {
        checkValve(ctx, function () {
          var drained = channel.publish(target.exchangeName, target.routingKey, zapper.bufferify(data), opts, function (err, ok) {
            connection.removeListener('error', unexpectedClosing);
            if (err) {
              onRejected(err);
              debugx.enabled && debugx('sendToExchange() failed: %s', JSON.stringify(err));
            } else {
              ok = ok || {};
              ok.drained = true;
              onResolved(ok);
              debugx.enabled && debugx('sendToExchange() is ok');
            }
          });
          if (!drained) {
            debugx.enabled && debugx('channel is overflowed, close the valve');
            ctx.state.valve && ctx.state.valve.set(false);
          }
        });
      } catch (exception) {
        connection.removeListener('error', unexpectedClosing);
        debugx.enabled && debugx('sendToExchange() throw exception: %s', JSON.stringify(exception));
        onRejected(exception);
      }
    });
  }).catch(function (err) {
    debugx.enabled && debugx('produce() - sendToExchange() failed, recall ...');
    ctx.state.connection = null;
    ctx.state.channel = null;
    return sendToExchange(ctx, data, opts, target);
  });
};

var getProducerState = function getProducerState(self) {
  self = self || this;
  self.producerState = self.producerState || {};
  self.producerState.guard = self.producerState.guard || locks.createMutex();
  self.producerState.valve = self.producerState.valve || locks.createCondVariable(true);
  if (self.config.exchangeMutex) {
    self.producerState.fence = self.producerState.fence || locks.createCondVariable(true);
  }
  if (self.config.exchangeQuota) {
    self.producerState.quota = self.producerState.quota || locks.createSemaphore(self.config.exchangeQuota);
  }
  return self.producerState;
};

var lockProducer = function lockProducer(self) {
  self = self || this;
  var producerState = getProducerState(self);
  var ticket = Promise.resolve(producerState);
  if (producerState.fence) {
    ticket = ticket.then(function (producerState) {
      return new Promise(function (onResolved, onRejected) {
        producerState.fence.wait(function conditionTest(value) {
          return value === true;
        }, function whenConditionMet() {
          onResolved(producerState);
        });
      });
    });
  }
  if (producerState.quota) {
    ticket = ticket.then(function (producerState) {
      return new Promise(function (onResolved, onRejected) {
        producerState.quota.wait(function whenResourceAvailable() {
          onResolved(producerState);
        });
      });
    });
  }
  return ticket;
};

var unlockProducer = function unlockProducer(self) {
  self = self || this;
  var producerState = getProducerState(self);
  if (producerState.quota) {
    producerState.quota.signal();
  }
};

Handler.prototype.ready = function () {
  var self = this;
  var sandbox = { config: self.config, state: getProducerState(self) };
  return getChannel(sandbox).then(function (ch) {
    var ok = assertExchange(sandbox, ch);
    if (hasInbox(self)) {
      ok = ok.then(function (eok) {
        return resolveInbox(self).then(function (ctx) {
          return assertQueue(ctx, ch).then(function (qok) {
            if (ctx.config.noBinding === true) return qok;
            return bindInbox(ch, qok, eok, self.config.routingKey);
          });
        });
      });
    }
    if (hasTrash(self)) {
      ok = ok.then(function (eok) {
        return resolveTrash(self).then(function (ctx) {
          return assertQueue(ctx, ch);
        });
      });
    }
    if (hasChain(self)) {
      ok = ok.then(function (eok) {
        return resolveChain(self).then(function (ctx) {
          return assertQueue(ctx, ch);
        });
      });
    }
    return ok;
  });
};

Handler.prototype.open = Handler.prototype.ready;
Handler.prototype.prepare = Handler.prototype.ready;

Handler.prototype.produce = function (data, opts, override) {
  var self = this;
  opts = opts || {};
  debugx.enabled && debugx('produce() an object to rabbitmq');
  return lockProducer(self).then(function (producerState) {
    var ctx = { config: self.config, state: producerState };
    return retrieveExchange(ctx, override).then(function (ref) {
      return sendToExchange(ctx, data, opts, ref);
    });
  }).finally(function () {
    unlockProducer(self);
  });
};

Handler.prototype.enqueue = Handler.prototype.produce;
Handler.prototype.publish = Handler.prototype.produce;

var getPipelineState = function getPipelineState(self) {
  self = self || this;
  self.pipelineState = self.pipelineState || {};
  self.pipelineState.guard = self.pipelineState.guard || locks.createMutex();
  self.pipelineState.valve = self.pipelineState.valve || locks.createCondVariable(true);
  if (self.config.exchangeMutex) {
    self.pipelineState.mutex = self.pipelineState.mutex || locks.createMutex();
  }
  return self.pipelineState;
};

var lockPipeline = function lockPipeline(self) {
  self = self || this;
  var producerState = getProducerState(self);
  if (producerState.fence) {
    producerState.fence.set(false);
  }
  var pipelineState = getPipelineState(self);
  return new Promise(function (onResolved, onRejected) {
    if (pipelineState.mutex) {
      pipelineState.mutex.lock(function () {
        onResolved(pipelineState);
      });
    } else {
      onResolved(pipelineState);
    }
  });
};

var unlockPipeline = function unlockPipeline(self) {
  self = self || this;
  var producerState = getProducerState(self);
  var pipelineState = getPipelineState(self);
  producerState.fence && producerState.fence.set(true);
  pipelineState.mutex && pipelineState.mutex.unlock();
};

var ProducerStream = function ProducerStream(context, options, target) {
  options = options || {};
  Writable.call(this, { objectMode: options.objectMode !== false });

  var count = 0;
  this.count = function () {
    return count;
  };

  this._write = function (chunk, encoding, callback) {
    debug2.enabled && debug2('ProducerStream._write() - sending #%s', count + 1);
    sendToExchange(context, chunk, options, target).then(function () {
      count++;
      debug2.enabled && debug2('ProducerStream._write() - #%s has been sent', count);
      callback();
    });
  };
};

util.inherits(ProducerStream, Writable);

Handler.prototype.exhaust = function (stream, opts, override) {
  var self = this;
  opts = opts || {};
  debugx.enabled && debugx('exhaust() data from a readable stream');
  if (!(stream instanceof Readable)) return Promise.reject({
    message: '[source] is not a readable stream'
  });
  return lockPipeline(self).then(function (pipelineState) {
    var ctx = { config: self.config, state: pipelineState };
    return retrieveExchange(ctx, override).then(function (ref) {
      var producerStream = new ProducerStream(ctx, opts, ref);
      return new Promise(function (resolved, rejected) {
        debug2.enabled && debug2('exhaust() - starting...');
        producerStream.on('error', function (err) {
          debug2.enabled && debug2('exhaust() - stream error: %s', JSON.stringify(err));
          rejected(err);
        });
        producerStream.on('finish', function () {
          debug2.enabled && debug2('exhaust() - stream end. Total: %s', producerStream.count());
          resolved();
        });
        stream.pipe(producerStream, { end: true });
      });
    });
  }).then(function (result) {
    debug2.enabled && debug2('exhaust() - finish successfully');
    return result;
  }).catch(function (error) {
    debug2.enabled && debug2('exhaust() - error occurred: %s', JSON.stringify(error));
    return Promise.reject(error);
  }).finally(function () {
    unlockPipeline(self);
  });
};

var getInboxState = function getInboxState(self) {
  self = self || this;
  self.inboxState = self.inboxState || {};
  self.inboxState.valve = self.inboxState.valve || locks.createCondVariable(true);
  return self.inboxState;
};

var hasInbox = function hasInbox(self) {
  self = self || this;
  return self.config && self.config.inbox && self.config.inbox.enabled !== false;
};

var resolveInbox = function resolveInbox(self) {
  self = self || this;
  if (!hasInbox(self)) return Promise.reject({ message: 'The inbox is unavailable' });
  return Promise.resolve({ name: 'enqueueInbox', config: self.config.inbox, state: getInboxState(self) });
};

var enqueueInbox = function enqueueInbox(self, data, opts) {
  return resolveInbox(self).then(function (ctx) {
    return sendToQueue(ctx, data, opts);
  });
};

var bindInbox = function bindInbox(ch, qok, eok, routingKey) {
  var ch_bindQueue = Promise.promisify(ch.bindQueue, { context: ch });
  return ch_bindQueue(qok.queue, eok.exchange, routingKey, {}).then(function (ok) {
    debugx.enabled && debugx('bindInbox() - queue: %s has been bound to %s', qok.queue, eok.exchange);
    return qok;
  });
};

Handler.prototype.checkInbox = function () {
  return resolveInbox(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return checkQueue(ctx, ch);
    });
  });
};

Handler.prototype.purgeInbox = function () {
  return resolveInbox(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return assertQueue(ctx, ch).then(function () {
        return purgeQueue(ctx, ch);
      });
    });
  });
};

Handler.prototype.consume = function (callback) {
  var self = this;

  assert.ok(zapper.isFunction(callback), 'callback should be a function');

  var ctx = { config: self.config.inbox, state: getInboxState(self) };
  ctx.state.count = ctx.state.count || 0;

  debugx.enabled && debugx('consume() - subscribe the Inbox queue');
  return getChannel(ctx).then(function (ch) {

    if (ctx.config.prefetch && ctx.config.prefetch >= 0) {
      debugx.enabled && debugx('consume() - set channel prefetch: %s', ctx.config.prefetch);
      ch.prefetch(ctx.config.prefetch, true);
    }

    var ok = assertQueue(ctx, ch, true);

    ok = ok.then(function (qok) {
      debugx.enabled && debugx('consume() - queue info: %s', JSON.stringify(qok));
      if (ctx.config.maxSubscribers && ctx.config.maxSubscribers <= qok.consumerCount) {
        var error = {
          consumerCount: qok.consumerCount,
          maxSubscribers: ctx.config.maxSubscribers,
          message: 'exceeding quota limits of subscribers'
        };
        debugx.enabled && debugx('consume() - Error: %s', JSON.stringify(error));
        return Promise.reject(error);
      }
      return qok;
    });

    ok = ok.then(function (qok) {
      if (ctx.config.noBinding === true) return qok;
      return assertExchange({ config: self.config, state: ctx.state }, ch).then(function (eok) {
        var ch_bindQueue = Promise.promisify(ch.bindQueue, { context: ch });
        return ch_bindQueue(qok.queue, eok.exchange, self.config.routingKey, {}).then(function () {
          debugx.enabled && debugx('consume() - queue: %s has been bound', qok.queue);
          return qok;
        });
      });
    });

    ok = ok.then(function (qok) {
      var ch_consume = Promise.promisify(ch.consume, { context: ch });
      return ch_consume(qok.queue, function (msg) {
        ctx.state.count++;
        debug0.enabled && debug0('consume() - received message: %s, fields: %s, amount: %s', msg.content, JSON.stringify(msg.fields), ctx.state.count);
        callback(msg.content, {
          fields: msg.fields,
          properties: msg.properties
        }, function done(err, msgz) {
          if (ctx.config.noAck !== true) {
            if (err) {
              var prom = void 0;
              var content = msg && msg.content && zapper.bufferify(msg.content) || null;
              var props = msg && msg.properties && lodash.clone(msg.properties) || {};
              var rcfg = self.config.trash;
              var headers = props.headers || {};
              headers[rcfg.redeliveredCountName] = (headers[rcfg.redeliveredCountName] || 0) + 1;
              if (headers[rcfg.redeliveredCountName] <= rcfg.redeliveredLimit) {
                debugx.enabled && debugx('consume() - enqueueInbox message');
                prom = enqueueInbox(self, content, props);
              } else {
                if (hasTrash(self)) {
                  debugx.enabled && debugx('consume() - enqueueTrash message');
                  prom = enqueueTrash(self, content, props);
                } else {
                  prom = Promise.resolve();
                }
              }
              prom.then(function () {
                ch.nack(msg, false, false);
              }).catch(function (err) {
                ch.nack(msg, false, true);
              });
            } else {
              if (hasChain(self)) {
                var _content = msgz && msgz.content && zapper.bufferify(msgz.content);
                _content = _content || msg && msg.content && zapper.bufferify(msg.content) || null;
                var _props = msgz && msgz.properties && lodash.clone(msgz.properties);
                _props = _props || msg && msg.properties && lodash.clone(msg.properties) || {};
                enqueueChain(self, _content, _props).then(function () {
                  ch.ack(msg);
                }).catch(function (err) {
                  ch.nack(msg, false, true);
                });
              } else {
                ch.ack(msg);
              }
            }
          }
          ctx.state.count--;
        });
      }, { noAck: ctx.config.noAck === true });
    });

    return ok.then(function (result) {
      debugx.enabled && debugx('consume() - consumerTag: %s. CTRL+C to exit', result.consumerTag);
      ctx.state.consumerRefs = ctx.state.consumerRefs || [];
      ctx.state.consumerRefs.push(result);
      return result;
    });
  });
};

Handler.prototype.process = Handler.prototype.consume;

var getChainState = function getChainState(self) {
  self = self || this;
  self.chainState = self.chainState || {};
  self.chainState.valve = self.chainState.valve || locks.createCondVariable(true);
  return self.chainState;
};

var hasChain = function hasChain(self) {
  self = self || this;
  return self.config && self.config.chain && self.config.chain.enabled !== false;
};

var resolveChain = function resolveChain(self) {
  self = self || this;
  if (!hasChain(self)) return Promise.reject({ message: 'The chain is unavailable' });
  return Promise.resolve({ name: 'enqueueChain', config: self.config.chain, state: getChainState(self) });
};

var enqueueChain = function enqueueChain(self, data, opts) {
  return resolveChain(self).then(function (ctx) {
    return sendToQueue(ctx, data, opts);
  });
};

Handler.prototype.checkChain = function () {
  return resolveChain(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return checkQueue(ctx, ch);
    });
  });
};

Handler.prototype.purgeChain = function () {
  return resolveChain(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return assertQueue(ctx, ch).then(function () {
        return purgeQueue(ctx, ch);
      });
    });
  });
};

var getTrashState = function getTrashState(self) {
  self = self || this;
  self.trashState = self.trashState || {};
  self.trashState.valve = self.trashState.valve || locks.createCondVariable(true);
  return self.trashState;
};

var hasTrash = function hasTrash(self) {
  self = self || this;
  return self.config && self.config.trash && self.config.trash.enabled !== false;
};

var resolveTrash = function resolveTrash(self) {
  self = self || this;
  if (!hasTrash(self)) return Promise.reject({ message: 'The trash is unavailable' });
  return Promise.resolve({ name: 'enqueueTrash', config: self.config.trash, state: getTrashState(self) });
};

var enqueueTrash = function enqueueTrash(self, data, opts) {
  return resolveTrash(self).then(function (ctx) {
    return sendToQueue(ctx, data, opts);
  });
};

Handler.prototype.checkTrash = function () {
  return resolveTrash(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return checkQueue(ctx, ch);
    });
  });
};

Handler.prototype.purgeTrash = function () {
  return resolveTrash(this).then(function (ctx) {
    return getChannel(ctx).then(function (ch) {
      return assertQueue(ctx, ch).then(function () {
        return purgeQueue(ctx, ch);
      });
    });
  });
};

Handler.prototype.recycle = function (callback) {
  var self = this;

  assert.ok(zapper.isFunction(callback), 'callback should be a function');

  return resolveTrash(self).then(function (ctx) {
    ctx.state.count = ctx.state.count || 0;

    debugx.enabled && debugx('recycle() - get an object from Trash');
    return getChannel(ctx).then(function (ch) {

      if (ctx.config.prefetch && ctx.config.prefetch >= 0) {
        debugx.enabled && debugx('recycle() - set channel prefetch: %s', ctx.config.prefetch);
        ch.prefetch(ctx.config.prefetch, true);
      }

      var ok = assertQueue(ctx, ch, true);

      ok = ok.then(function (qok) {
        debugx.enabled && debugx('recycle() - queue info: %s', JSON.stringify(qok));
        if (ctx.config.maxSubscribers && ctx.config.maxSubscribers <= qok.consumerCount) {
          var error = {
            consumerCount: qok.consumerCount,
            maxSubscribers: ctx.config.maxSubscribers,
            message: 'exceeding quota limits of subscribers'
          };
          debugx.enabled && debugx('recycle() - Error: %s', JSON.stringify(error));
          return Promise.reject(error);
        }
        return qok;
      });

      ok = ok.then(function (qok) {
        return Promise.promisify(ch.consume, { context: ch })(qok.queue, function (msg) {
          ctx.state.count++;
          debugx.enabled && debugx('recycle() - received message: %s, fields: %s, amount: %s', msg.content, JSON.stringify(msg.fields), ctx.state.count);
          callback(msg.content, {
            fields: msg.fields,
            properties: msg.properties
          }, function done(err) {
            if (ctx.config.noAck !== true) {
              err ? ch.nack(msg) : ch.ack(msg);
            }
            ctx.state.count--;
          });
        }, { noAck: ctx.config.noAck === true });
      });

      return ok.then(function (result) {
        debugx.enabled && debugx('recycle() - consumerTag: %s. CTRL+C to exit', result.consumerTag);
        ctx.state.consumerRefs = ctx.state.consumerRefs || [];
        ctx.state.consumerRefs.push(result);
        return result;
      });
    });
  });
};

var examineGarbage = function examineGarbage(self) {
  self = self || this;
  return resolveTrash(self).then(function (ctx) {
    if (ctx.state.garbage) return ctx.state.garbage;
    return getChannel(ctx).then(function (ch) {
      return assertQueue(ctx, ch).then(function (qok) {
        return Promise.promisify(ch.get, { context: ch })(qok.queue, {});
      }).then(function (msgOrFalse) {
        debugx.enabled && debugx('examineGarbage() - msg: %s', JSON.stringify(msgOrFalse));
        if (msgOrFalse !== false) ctx.state.garbage = msgOrFalse;
        return ctx.state.garbage;
      });
    });
  });
};

var garbageAction = {};

var discardGarbage = garbageAction['discard'] = function (self) {
  self = self || this;
  return resolveTrash(self).then(function (ctx) {
    if (!ctx.state.garbage) return false;
    return getChannel(ctx).then(function (ch) {
      debugx.enabled && debugx('discardGarbage() - nack()');
      ch.nack(ctx.state.garbage, false, false);
      ctx.state.garbage = undefined;
      return true;
    });
  });
};

var restoreGarbage = garbageAction['restore'] = function (self) {
  self = self || this;
  return resolveTrash(self).then(function (ctx) {
    if (!ctx.state.garbage) return false;
    return getChannel(ctx).then(function (ch) {
      debugx.enabled && debugx('restoreGarbage() - nack()');
      ch.nack(ctx.state.garbage);
      ctx.state.garbage = undefined;
      return true;
    });
  });
};

var recoverGarbage = garbageAction['recover'] = function (self) {
  self = self || this;
  return resolveTrash(self).then(function (ctx) {
    if (!ctx.state.garbage) return false;
    return getChannel(ctx).then(function (ch) {
      var msg = ctx.state.garbage;
      return enqueueInbox(self, msg.content, msg.properties).then(function (result) {
        debugx.enabled && debugx('recoverGarbage() - ack()');
        ch.ack(ctx.state.garbage);
        ctx.state.garbage = undefined;
        return true;
      });
    });
  });
};

var requeueGarbage = garbageAction['requeue'] = function (self) {
  self = self || this;
  return resolveTrash(self).then(function (ctx) {
    if (!ctx.state.garbage) return false;
    return getChannel(ctx).then(function (ch) {
      var msg = ctx.state.garbage;
      return enqueueTrash(self, msg.content, msg.properties).then(function (result) {
        debugx.enabled && debugx('requeueGarbage() - ack()');
        ch.ack(ctx.state.garbage);
        ctx.state.garbage = undefined;
        return true;
      });
    });
  });
};

Handler.prototype.examine = function (callback) {
  var self = this;
  var result = { obtained: 0 };
  result.callback = zapper.isFunction(callback);
  return examineGarbage(self).then(function (msg) {
    if (msg) {
      result.obtained = 1;
      if (!result.callback) return result;
      return new Promise(function (resolved, rejected) {
        var copied = lodash.pick(msg, ['content', 'fields', 'properties']);
        callback(copied, function update(action, data) {
          if (zapper.isObject(data)) {
            msg.content = zapper.bufferify(data.content);
            if (zapper.isObject(data.properties)) {
              lodash.merge(msg.properties, data.properties);
            }
          }
          if (zapper.isFunction(garbageAction[action])) {
            result.nextAction = true;
            garbageAction[action](self).then(function () {
              resolved(result);
            }).catch(function (err) {
              result.actionError = err;
              rejected(result);
            });
          } else {
            result.nextAction = false;
            resolved(result);
          }
        });
      });
    } else {
      return result;
    }
  });
};

Handler.prototype.close = function () {
  var self = this;
  return Promise.resolve().then(function () {
    return stopPublisher({ state: getProducerState(self) });
  }).then(function () {
    return stopPublisher({ state: getPipelineState(self) });
  }).then(function () {
    return stopSubscriber({ state: getInboxState(self) });
  }).then(function () {
    return stopSubscriber({ state: getTrashState(self) });
  }).then(function () {
    return stopSubscriber({ state: getChainState(self) });
  });
};

Handler.prototype.destroy = Handler.prototype.close;

var stopPublisher = function stopPublisher(self) {
  self = self || this;
  return Promise.resolve().then(function () {
    if (self.state.channel) {
      var ch_close = Promise.promisify(self.state.channel.close, {
        context: self.state.channel
      });
      return ch_close().then(function (ok) {
        delete self.state.channel;
        debugx.enabled && debugx('destroy() - channel is closed: %s', JSON.stringify(ok));
        return true;
      });
    }
    return true;
  }).then(function () {
    if (self.state.connection) {
      var ch_close = Promise.promisify(self.state.connection.close, {
        context: self.state.connection
      });
      return ch_close().then(function (ok) {
        delete self.state.connection;
        debugx.enabled && debugx('destroy() - connection is closed: %s', JSON.stringify(ok));
        return true;
      });
    }
    return true;
  });
};

var stopSubscriber = function stopSubscriber(self) {
  self = self || this;
  return Promise.resolve().then(function () {
    if (self.state.channel && self.state.consumerRefs && self.state.consumerRefs.length > 0) {
      debugx.enabled && debugx('stopSubscriber() - consumers cancellation has been invoked');
      var ch_cancel = Promise.promisify(self.state.channel.cancel, {
        context: self.state.channel
      });
      return Promise.mapSeries(self.state.consumerRefs, function (consumerRef) {
        return ch_cancel(consumerRef.consumerTag).then(function (ok) {
          debugx.enabled && debugx('stopSubscriber() - consumer is cancelled: %s', JSON.stringify(ok));
          return ok;
        });
      });
    }
    return true;
  }).then(function () {
    if (self.state.channel) {
      var ch_close = Promise.promisify(self.state.channel.close, {
        context: self.state.channel
      });
      return ch_close().then(function (ok) {
        delete self.state.channel;
        debugx.enabled && debugx('stopSubscriber() - channel is closed: %s', JSON.stringify(ok));
        return true;
      });
    }
    return true;
  }).then(function () {
    if (self.state.connection) {
      var ch_close = Promise.promisify(self.state.connection.close, {
        context: self.state.connection
      });
      return ch_close().then(function (ok) {
        delete self.state.connection;
        debugx.enabled && debugx('stopSubscriber() - connection is closed: %s', JSON.stringify(ok));
        return true;
      });
    }
    return true;
  });
};

module.exports = Handler;